; $Id: hdf_sd__define.pro,v 1.1 2012/03/13 16:03:55 awalther Exp $
;+
; NAME:
;	hdf_sd
;
; FILENAME:
;	hdf_sd__define.pro
;
; PURPOSE:
;	This program creates a kind of read-only dictionary from
;	the scientific datasets of an hdfv4 file or from a netcdf file.
;	The data and the attributes can be accessed by their names
;	(see methods). Some additional methods are implemented.
;
; AUTHOR:
;	Rene Preusker
;	Freie Universität Berlin
;	e-mail: rene.preusker@fu-berlin.de
;
; CATEGORY:
;	data 
;
; CALLING SEQUENCE:
;	ohdf=obj_new('hdf_sd',hdf_filename)
;
; OPTIONAL INPUTS:
;	none
;
; RESTRICTIONS:
;	It is read only, you can not create or extent hdf files 
;
;
; PUBLIC METHODS:
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; FUNCTION HDF_SD::HAS_ATT, att_name, found=found
;	returns true if true
;
; KEYWORDS: found
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; FUNCTION HDF_SD::HAS_SDS, sds_name, found=found
;	returns true if true
;
; KEYWORDS: found
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; FUNCTION HDF_SD::GET_ATT_NAMES
;	returns string vector of all global atributes
;
; KEYWORDS: found
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; FUNCTION HDF_SD::GET_SDS_NAMES
;	returns string vector of all scientific data sets
;
; KEYWORDS: found
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; FUNCTION HDF_SD::GET_SDS_ATT_NAMES, sds_name, found=found
;	returns string vector of all atributes belonging to sds_name
;
; KEYWORDS: found
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; FUNCTION HDF_SD::GET_ATT, att_name, found=found
;	returns value of global attribute att_name
;
; KEYWORDS: found
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; FUNCTION HDF_SD::GET_SDS_ATT, sds_name, att_name, found=found
;	returns value of attribute att_name belonging to sds_name
;
; KEYWORDS: found
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; FUNCTION HDF_SD::GET_SDS_INFO, sds_name, found=found
;	returns a structure of all information tags belonging to sds_name
;
; KEYWORDS: found
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; FUNCTION HDF_SD::GET_SDS, sds_name, found=found  		$
;		,offset=offset,count=count,stride=stride	$
;		,no_derefrence=no_dereference
;	returns the data of sds_name
;
; KEYWORDS: found, offset,count, stride, no_dereference       
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
;PRIVATE Methods
; (don't exist in idl) Here they start with "_"
;
; EXAMPLE: 
;	(use hdf_sd_test ..)
; 	o=obj_new("hdf_sd","/people/rene/idltools/MYD06.hdf")
;	o=obj_new("hdf_data",a)
; 	print,o->has_sds("longitude")
; 	print,o->has_sds("Longitude")
; 	print,o->has_att("Longitude")
; 	print,o->has_att("title")
; 	help,o->get_att("Longitude")
; 	print,o->get_att("title")
; 	print,o->get_att_names()
; 	print,o->get_sds_names()
; 	print,o->get_sds_info("longitude")
; 	print,(o->get_sds_info("Longitude")).dims
; 	help,o->get_sds_info("Longitude"),/str
; 	print,o->get_sds_att_names("Longitude")
; 	print,o->get_sds_att("Longitude","units")
; 	help,o->get_sds("Longitude")
; 	obj_destroy,o
;
;
; MODIFICATION HISTORY:
;	Written by: Rene Preusker, 19.11.2010
;
; LICENSE
; Copyright (C) 2010, Freie Universitaet Berlin
; This software may be used, copied, or redistributed as long
; as it is not sold and this copyright notice is reproduced on
; each copy made.  This routine is provided as is without any
; express or implied warranties whatsoever.
;
;-
pro hdf_sd::_error_msg,which,what
        print,"Error in "+which+": "+what
end
function hdf_sd::_is_netcdf,file
        dum=bytarr(3)
        openr,lun,file,/get_lun
        readu,lun,dum
        free_lun,lun
	; "CDF"
        if total(dum eq [67b,68b,70b]) eq 3b then ret=1 else ret=0
        return,ret
end
function hdf_sd::_name_to_index,vec,name,found=found
	if size(name,/type) ne 7 then begin
		found=0
		wo=-1
		self->_error_msg,"Warning","name not a string"
	endif else begin	
        	wo=where(vec eq name,count)
		found = count eq 1
	endelse
	return,wo
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;End private routines 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Begin information routines 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
function hdf_sd::has_att,att_name 
; is true if .... 
	for i=0L,self.numats-1L do begin
		if (*self.atrstr)[i].name eq att_name then return,1b
	endfor	
	return,0b
end
function hdf_sd::has_sds,sds_name
; is true if .... 
	for i=0L,self.numsds-1L do begin
		if (*self.sdsstr)[i].name eq sds_name then return,1b
	endfor	
	return,0b
end
function hdf_sd::get_att_names
; returns a string vector of all global attribute names 
	return,(*self.atrstr).name
end
function hdf_sd::get_sds_names
; returns a string vector of all sds names 
	return,(*self.sdsstr).name
end
function hdf_sd::get_sds_att_names,name,found=found
; returns a string vector of all attribute names belonging to sds_name 
	index=self->_name_to_index((*self.sdsstr).name,name,found=found)
	if found then begin
		dumstr=*((*self.sdsstr)[index].attr)
		return,dumstr.name
	endif
	return,0b
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;End information routines 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;Begin get_attribute routines 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
function hdf_sd::get_att,att_name,found=found
; returns the value of the global attribute "name" 
	index=self->_name_to_index((*self.atrstr).name,att_name,found=found)
	if found then begin
		return,*(*self.atrstr)[index].value
	endif
end
function hdf_sd::get_sds_att,sds_name,att_name,found=found
; returns the value of the attribute "a_name" of sds "s_name"
	found=0b
	index=self->_name_to_index((*self.sdsstr).name,sds_name,found=sfound)
	if sfound then begin
		dumstr=*((*self.sdsstr)[index].attr)
		andex=self->_name_to_index(dumstr.name,att_name,found=afound)
		found = sfound and afound
		if found then return,*(dumstr[andex].value)
	endif
	return,0b
end
function hdf_sd::get_sds_info,sds_name,found=found
; returns a sructure  of all "sds_info" belonging to name 
	index=self->_name_to_index((*self.sdsstr).name,sds_name,found=found)
	if found then begin
		return,*((*self.sdsstr)[index].info)
	endif
	return,0b 
end
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;End get_attribute routines 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;The get data routine 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
function hdf_sd::get_sds,sds_name,found=found			$
			,offset=offset,count=count,stride=stride	$
			,no_derefrence=no_dereference

	index=self->_name_to_index((*self.sdsstr).name,sds_name,found=found)
	
	if found then begin
		hdf_sd_getdata,(*self.sdsstr)[index].sdid, data,count=count,start=offset,stride=stride
		if keyword_set(no_dereference)then return,ptr_new(data)
		return,data
	endif else return,0b
end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;The rest 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
pro hdf_sd::cleanup
	; usually clean up pointers
	; and close file handles
	heap_free, self.sdsstr,/ptr	;,/verb
	heap_free, self.atrstr,/ptr	;,/verb
	hdf_sd_end, self.file_handle
end


function hdf_sd::init,filename,debug=debug
	if ~keyword_set(debug) then begin
		on_error,2
		self.debug=1
	endif else self.debug=0	

	if ~(size(filename,/type) gt 0) then begin
		self->_error_msg,"init","filename not defined"
		return,0b
	endif
	if ~(size(filename,/type) eq 7) then begin
		self->_error_msg,"init","filename not a string"
		return,0b
	endif
	if ~ (n_elements(filename) eq 1) then begin
		self->_error_msg,"init","filename has more than 1 element"
		return,0b
	endif
	if ~(file_test(filename,/read) eq 1) then begin
		self->_error_msg,"init","file not found or readable"
		return,0b
	endif
	if ~( hdf_ishdf(filename) or self->_is_netcdf(filename) ) then begin
		self->_error_msg,"init","file not an hdf v4 or ncdf"
		return,0b
	endif
	self.filename=filename
	self.file_handle=hdf_sd_start(filename, /read)
	; first get basic info
	hdf_sd_fileinfo, self.file_handle, numsds, numats
	; put them into self
	self.numsds=numsds
	self.numats=numats
    
	IF numsds eq 0 then begin
	  self->_error_msg,"init","no sds in file"
	  return,0B
	endif
	
	; get basic info for all sds
 	sdsstr=replicate({name:"",sdid:0UL,natt:0UL,attr:ptr_new(),info:ptr_new()},numsds)
	for i_numsds=0L,numsds-1L do begin
		qq=!quiet
		!quiet=1
		sdsid = hdf_sd_select(self.file_handle,i_numsds)		
		hdf_sd_getinfo, sdsid,caldata=caldata,coordsys=coordsys		$
			,dims=dims,fill=fill,format=format			$
			,hdf_type=hdf_type,label=label,name=sds_name		$
			,natts=natts,ndims=ndims				$
			,range=range,type=type,unit=unit 
		hdf_sd_endaccess,sdsid
		!quiet=qq
		sdsstr[i_numsds].name=sds_name
		sdsstr[i_numsds].sdid=sdsid
		sdsstr[i_numsds].natt=natts
		IF natts GT 0 THEN BEGIN
		  atrstr=replicate({name:"",value:ptr_new()},natts)
		  for i_numats=0L,natts-1L do begin
	   		hdf_sd_attrinfo, sdsid, i_numats, name=att_name, data=att_data
			atrstr[i_numats].name=att_name
			atrstr[i_numats].value=ptr_new(att_data)
		  endfor
		  sdsstr[i_numsds].attr=ptr_new(atrstr)
		  infstr={name:sds_name}
		  if size(caldata, /type) gt 0 then infstr=create_struct(infstr,'caldata' ,caldata)
		  if size(coordsys,/type) gt 0 then infstr=create_struct(infstr,'coordsys',coordsys)
		  if size(dims,    /type) gt 0 then infstr=create_struct(infstr,'dims'    ,dims)
		  if size(fill,    /type) gt 0 then infstr=create_struct(infstr,'fill'    ,fill)
		  if size(format,  /type) gt 0 then infstr=create_struct(infstr,'format'  ,format)
		  if size(hdf_type,/type) gt 0 then infstr=create_struct(infstr,'hdf_type',hdf_type)
		  if size(label,   /type) gt 0 then infstr=create_struct(infstr,'label'   ,label)
		  if size(natts,   /type) gt 0 then infstr=create_struct(infstr,'natts'   ,natts)
		  if size(ndims,   /type) gt 0 then infstr=create_struct(infstr,'ndims'   ,ndims)
		  if size(range,   /type) gt 0 then infstr=create_struct(infstr,'range'   ,range)
		  if size(type,    /type) gt 0 then infstr=create_struct(infstr,'type'    ,type)
		  if size(unit,    /type) gt 0 then infstr=create_struct(infstr,'unit'    ,unit)
;		infstr={caldata:caldata,coordsys:coordsys		$
;			,dims:dims,fill:fill,format:format		$
;			,hdf_type:hdf_type,label:label,name:sds_name	$
;			,natts:natts,ndims:ndims,range:range		$
;			,type:type,unit:unit}
		  sdsstr[i_numsds].info=ptr_new(infstr)
      ENDIF
	endfor
	; put it into self
	self.sdsstr=ptr_new(sdsstr)

	; get all global attributes
	atrstr=replicate({name:"",value:ptr_new()},numats)
	for i_numats=0L,self.numats-1L do begin
		hdf_sd_attrinfo,self.file_handle,i_numats,name=att_name,data=att_data
		atrstr[i_numats].name=att_name
		atrstr[i_numats].value=ptr_new(att_data)
	endfor
	; put it into self
	self.atrstr=ptr_new(atrstr)
	return,1b	
end
pro hdf_sd__define
	bla={hdf_sd,debug:0b,file_handle:0UL	$
		     	,numsds:0UL		$
			,numats:0UL		$
			,filename:""		$
			,sdsstr:ptr_new()	$
			,atrstr:ptr_new()	$
			}
end

pro hdf_sd_test
	o=obj_new("hdf_sd","/people/rene/idltools/MYD06.hdf")
;	o=obj_new("hdf_data",a)
	print,o->has_sds("longitude")
	print,o->has_sds("Longitude")
	print,o->has_att("Longitude")
	print,o->has_att("title")
	help,o->get_att("Longitude")
	print,o->get_att("title")
	print,o->get_att_names()
	print,o->get_sds_names()
	print,o->get_sds_info("longitude")
	print,(o->get_sds_info("Longitude")).dims
	help,o->get_sds_info("Longitude"),/str
	print,o->get_sds_att_names("Longitude")
	print,o->get_sds_att("Longitude","units")
	help,o->get_sds("Longitude")
	print,o->get_sds_att(a,"")

	obj_destroy,o
	
end

