;%$Id: read_standard_hdf.pro,v 1.30 2015/06/10 15:14:58 heidinger Exp $
;-----------------------------------------------------------------
;--- function that will read hdf sds that is scaled
;--- using the standard CF Compliant attributes
;--- 
;--- This routine only handle unscaled or linearly scaled data
;----
;--- data is unscaled as follows:
;--- unscaled = scaled * scale_factor + add_offset 
;---
;--- where the scaled data is equal to the fill value, the 
;--- unscaled data is set to the missing data
;---
;--- The missing value is assumed to be -999.0
;--- which is the CLAVR-x default.
;---
;--- if the fill value is missing, it is set to smallest value
;--- allowed for the data type, this is also the CLAVR-x default
;---
;--- if an unscaled sds has a missing value of -888.0, it is 
;--- treated as bitmask and no missing value is applied
;---
;--- Note, IDL does not support signed one byte integers, IDL
;--- converts them to bytes (0-255).  These are converted to
;--- two byte integers and value above 127 are corrected.
;------------------------------------------------------------------
function read_standard_hdf, file_name, sds_name, quiet=quiet

;on_error, 2

;--- set default missing values for use when the reading of them fails
default_missing_value_r4 = -999.0
default_missing_value = -999.0

formerquietstatus=!QUIET
if (keyword_set(quiet) ne 0)then !QUIET=1

;--- open the file
sd_id = hdf_sd_start(file_name) 

;--- handle a failure of the opening
if (sd_id eq -1) then begin
   if (keyword_set(quiet) eq 0) then print, 'could not open file named ', file_name
   hdf_sd_end,sd_id
   return, -1
endif

;--- find sds in the file
index = hdf_sd_nametoindex(sd_id,sds_name)

;-- handle failure of finding sds
if (index eq -1) then begin
   if (keyword_set(quiet) eq 0) then print, 'could not open sds named ', sds_name
   hdf_sd_end,sd_id
   return, -1
endif

;--- open the sds
sds_id = hdf_sd_select(sd_id,index)

;--- handle failure of opening the sds
if (sds_id eq -1) then begin
   if (keyword_set(quiet) eq 0) then print, 'could not open sds named ', sds_name
   hdf_sd_end,sd_id
   return, -1
endif

;--- get information of this file
hdf_sd_getinfo,sds_id,hdf_type=data_type

;--- read the data
hdf_sd_getdata,sds_id,data

;--- read scaling attribute
attr_index = hdf_sd_attrfind(sds_id,'SCALED')
if (attr_index eq -1) then begin
    ;--- look for scale_factor, if present assume linear scaling
    attr_index = hdf_sd_attrfind(sds_id,'scale_factor')
    if (attr_index ne -1) then  begin
        scaling_method = [1]
    endif else begin
        scaling_method = [0]
         print, 'inappropriate scaling method, assuming data is unscaled'
    endelse
endif else begin
  hdf_sd_attrinfo,sds_id,attr_index,DATA=scaling_method
endelse

;--- read scaling attribute
missing_value = [default_missing_value] 

;--- test scaling_method
if (scaling_method[0] ne 0 and scaling_method[0] ne 1) then begin
   scaling_method[0] = 0
   print, 'inappropriate scaling method, assuming data is unscaled'
endif

;--- determine data type and account for byte / int1 mismatch
if (data_type[0] eq 'DFNT_INT8') then begin
  data = long(data)
  index = where(data ge 128L,nindex)
  if (nindex gt 0) then data[index] = data[index] - 256L
endif

;--- read scale factor
if (scaling_method[0] ge 1) then begin
  attr_index = hdf_sd_attrfind(sds_id,'scale_factor')
  if (attr_index eq -1) then begin
   attr_index = hdf_sd_attrfind(sds_id,'SCALE_FACTOR')
  endif
  if (attr_index eq -1 and keyword_set(quiet) eq 0) then print, 'could not read scale_factor attribute for ', sds_name
endif else begin
  attr_index = -1
endelse

if (attr_index eq -1) then begin
   scale_factor = [1.0]
endif else begin
  hdf_sd_attrinfo,sds_id,attr_index,DATA=scale_factor
endelse

;--- read add offset
if (scaling_method[0] ge 1) then begin
 attr_index = hdf_sd_attrfind(sds_id,'add_offset')
 if (attr_index eq -1) then begin
   attr_index = hdf_sd_attrfind(sds_id,'ADD_OFFSET')
 endif
 if (attr_index eq -1 and keyword_set(quiet) eq 0) then print, 'could not read add_offset attribute for ', sds_name
endif else begin
  attr_index = -1
endelse

if (attr_index eq -1) then begin
  offset = [0.0]
endif else begin
  hdf_sd_attrinfo,sds_id,attr_index,DATA=add_offset
endelse

;--- read fill value
fill_value = [0L]
if (scaling_method[0] eq 1) then begin
 attr_index = hdf_sd_attrfind(sds_id,'_FillValue')
 if (attr_index eq -1) then begin
   attr_index = hdf_sd_attrfind(sds_id,'_FILLVALUE')
 endif
 if (attr_index eq -1) then begin
    if (data_type[0] eq 'DFNT_INT8') then fill_value = [-128]
    if (data_type[0] eq 'DFNT_INT16') then fill_value = [-32768]
 endif else begin
  hdf_sd_attrinfo,sds_id,attr_index,DATA=fill_value
    if (data_type[0] eq 'DFNT_INT8') then  begin
       fill_value_temp = long(fill_value[0])
       if (fill_value_temp ge 128) then fill_value_temp = fill_value_temp - 256L
       fill_value = [fill_value_temp]
    endif
 endelse
endif

;--- close the sds
hdf_sd_endaccess,sds_id

;--- close the file
hdf_sd_end,sd_id

!QUIET=formerquietstatus

;--- unscale data if necessary
if (scaling_method[0] eq 1) then begin
   nbad = 0
   index = where(data eq fill_value[0],nbad)
   if (scaling_method[0] eq 1) then begin
      data = temporary(data)*scale_factor[0] + add_offset[0]
   endif
   if (nbad gt 0)  then data[index] = default_missing_value_r4
endif
if (scaling_method[0] eq 0 and missing_value[0] ne -888.0) then begin
   nbad = 0
   index = where(data eq missing_value[0],nbad)
   if (nbad gt 0)  then data[index] = default_missing_value_r4
endif

return,data

end
